/*****************************************************************************/
/*                      WiNRADiO G65DDC EXCALIBUR SIGMA                      */
/*                          Coherent mode SDK DEMO                           */
/*                        (c) Copyright RADIXON 2023                         */
/*****************************************************************************/

#include "u_devicepanel.h"
#include "u_mainwindow.h"

enum
{
    ID_DDC1_SPECTRUM=wxID_HIGHEST,
    ID_PHASE_SHIFT,
    ID_IF_SPECTRUM
};

BEGIN_EVENT_TABLE(TDevicePanel, wxPanel)
    EVT_SPECTRUM_TUNE(ID_IF_SPECTRUM,TDevicePanel::OnSpectrumTune)
    EVT_NUMERICEDIT(ID_PHASE_SHIFT,TDevicePanel::OnNumericEditChanged)
END_EVENT_TABLE()




TDevicePanel::TDevicePanel(TMainWindow *MainWindow,ICohG65DDCDeviceSet *DeviceSet,uint32_t DeviceIndex):
    wxPanel(MainWindow,wxID_ANY,wxPoint(0,0),wxSize(100,100),wxCLIP_CHILDREN|wxBORDER_NONE),
    m_DDC1FFTOverlappingBuffer(4*FFT_SIZE)
{
 G65DDC_DEVICE_INFO DeviceInfo;
 wxString s;
 wxSizer *MainSizer,*Sizer;
 wxSize Size;

    m_MainWindow=MainWindow;

    m_DeviceSet=DeviceSet;
    m_DeviceIndex=DeviceIndex;

    m_DeviceSet->GetDeviceInfo(m_DeviceIndex,&m_DeviceInfo);

    m_DDC1FFTOverlappingBuffer.SetSize(FFT_SIZE);

    GetNormalizedWindowCoeffs(m_DDC1FFTCoeffs,FFT_SIZE,1);

    GetNormalizedWindowCoeffs(m_ADCFFTCoeffs,FFT_SIZE,1/32768.0);

    m_ADCSpectrumCenterFrequency=0;

    MainSizer=new wxBoxSizer(wxVERTICAL);
    SetSizer(MainSizer);


    Sizer=new wxBoxSizer(wxHORIZONTAL);
    MainSizer->Add(Sizer,0,wxEXPAND|wxTOP|wxBOTTOM,4);

    DeviceSet->GetDeviceInfo(DeviceIndex,&DeviceInfo);

    s.sprintf(wxT("Device[%u] - %s"),DeviceIndex,wxString(DeviceInfo.SerialNumber,wxConvLocal).c_str());

    m_LSerNum=new wxStaticText(this,wxID_ANY,s);
    Sizer->Add(m_LSerNum,0,wxALIGN_LEFT|wxALIGN_CENTRE_VERTICAL);

    m_LPhaseShift=new wxStaticText(this,wxID_ANY,wxT("Phase shift"));
    Sizer->Add(m_LPhaseShift,0,wxALIGN_LEFT|wxALIGN_CENTRE_VERTICAL|wxLEFT,8);

    m_EPhaseShift=new TFloatEdit(this,ID_PHASE_SHIFT,0,-1800,1800,1,wxDefaultPosition,wxSize(50,-1));

    Size=m_EPhaseShift->GetTextExtent(wxT("999.9"));
    Size=m_EPhaseShift->GetSizeFromTextSize(Size.x+20);

    m_EPhaseShift->SetInitialSize(Size);

    Sizer->Add(m_EPhaseShift,0,wxALIGN_LEFT|wxALIGN_CENTRE_VERTICAL|wxLEFT,4);

    Sizer->AddStretchSpacer();

    m_LState=new wxStaticText(this,wxID_ANY,wxT("N/A"));
    Sizer->Add(m_LState,0,wxALIGN_CENTER_VERTICAL|wxALIGN_RIGHT);

    m_DDC1Spectrum=new TSpectrum(this,ID_DDC1_SPECTRUM);
    MainSizer->Add(m_DDC1Spectrum,1,wxEXPAND);

    m_DDC1Spectrum->SetBand(SpectrumBand_None,false);

    s.sprintf(wxT("DDC1 spectrum / device %u"),m_DeviceIndex);

    m_DDC1Spectrum->SetText(s);

    Sizer->AddSpacer(4);


    m_ADCSpectrum=new TSpectrum(this,ID_IF_SPECTRUM);
    MainSizer->Add(m_ADCSpectrum,1,wxEXPAND);

    m_ADCSpectrum->SetBand(SpectrumBand_BothSides,false);

    s.sprintf(wxT("IF spectrum / device %u"),m_DeviceIndex);

    m_ADCSpectrum->SetText(s);

    UpdateRange();
    UpdateDDC1();
    UpdateDDC1Frequency();

    UpdateControls();
}


TDevicePanel::~TDevicePanel(void)
{
}


void TDevicePanel::UpdateControls(void)
{
 int Power;
 bool e;

    m_DeviceSet->GetPower(&Power);

    e=Power!=FALSE;

    m_ADCSpectrum->Enable(e);
    m_DDC1Spectrum->Enable(e);
    m_LState->Enable(e);

    m_LPhaseShift->Enable(e);
    m_EPhaseShift->Enable(e);

    if(!e)
    {
        m_DDC1Spectrum->Clear();
        m_LState->SetForegroundColour(m_LPhaseShift->GetForegroundColour());
        m_LState->SetLabel(wxT("N/A"));
    }
}

void TDevicePanel::UpdateDDC1FFTCompensation(void)
{
 int32_t Frequency;

    m_DeviceSet->GetDDC1Frequency(&Frequency);

    m_DDC1FFTMutex.Lock();

    m_DeviceSet->GetSpectrumCompensation(m_DeviceIndex,Frequency,m_DDC1Info.SampleRate,m_DDC1FFTCompensation,FFT_SIZE);

    m_DDC1FFTMutex.Unlock();
}

void TDevicePanel::UpdateRange(void)
{
 uint32_t Range;

    m_DeviceSet->GetRange(&Range);

    m_ADCFFTMutex.Lock();

    m_ADCSpectrumBW=m_DeviceInfo.Ranges[Range].MaxFrequency-m_DeviceInfo.Ranges[Range].MinFrequency;
    m_RangeCenterFrequency=(m_DeviceInfo.Ranges[Range].MaxFrequency+m_DeviceInfo.Ranges[Range].MinFrequency)/2;

    m_ADCSpectrum->SetRange(m_ADCSpectrumBW);
    m_ADCSpectrum->SetCenterFrequency(m_RangeCenterFrequency);

    m_ADCSpectrumCenterFrequency=0;  //this ensures compensation data update in the ADC snapshot callback

    m_ADCFFTMutex.Unlock();
}

void TDevicePanel::UpdateDDC1(void)
{
    m_DeviceSet->GetDDC1(NULL,&m_DDC1Info);

    m_DDC1Spectrum->SetRange(m_DDC1Info.Bandwidth);

    m_ADCSpectrum->SetBandWidth(m_DDC1Info.Bandwidth);

    m_DDC1FFTOverlappingBuffer.Reset();
}

void TDevicePanel::UpdateDDC1Frequency(void)
{
 int32_t Frequency;

    m_DeviceSet->GetDDC1Frequency(&Frequency);

    m_DDC1Spectrum->SetCenterFrequency(Frequency);

    m_ADCSpectrum->SetBandFrequency(Frequency);

    UpdateDDC1FFTCompensation();
}

void TDevicePanel::DDC1StreamCallback(const void *Buffer,uint32_t Count,uint32_t BitsPerSample)
{
 const float *OverlBuffer;
 float *FFTBuffer,*FFTCoeffs,m,TempI,TempQ;
 uint32_t i;

    if(BitsPerSample==16)
    {
        m_DDC1FFTOverlappingBuffer.AddSamples_16((int16_t*)Buffer,Count);
    }
    else
    {
        m_DDC1FFTOverlappingBuffer.AddSamples_32((int32_t*)Buffer,Count);
    }

    OverlBuffer=m_DDC1FFTOverlappingBuffer.Get();

    if(OverlBuffer)
    {
        FFTBuffer=m_DDC1FFTBuffer;
        FFTCoeffs=m_DDC1FFTCoeffs;

        for(i=0;i<FFT_SIZE;i++)
        {
            FFTBuffer[i*2]=OverlBuffer[i*2]*FFTCoeffs[i];
            FFTBuffer[i*2+1]=OverlBuffer[i*2+1]*FFTCoeffs[i];
        }

        FFT(FFTBuffer,FFT_SIZE);

        //swap lower and upper half of the FFT result
        for(i=0;i<FFT_SIZE/2;i++)
        {
            TempI=FFTBuffer[i*2];
            TempQ=FFTBuffer[i*2+1];

            FFTBuffer[i*2]=FFTBuffer[(i+FFT_SIZE/2)*2];
            FFTBuffer[i*2+1]=FFTBuffer[(i+FFT_SIZE/2)*2+1];

            FFTBuffer[(i+FFT_SIZE/2)*2]=TempI;
            FFTBuffer[(i+FFT_SIZE/2)*2+1]=TempQ;
        }

        m_DDC1FFTMutex.Lock();

        for(i=0;i<FFT_SIZE;i++)
        {
            m=FFTBuffer[i*2]*FFTBuffer[i*2]+FFTBuffer[i*2+1]*FFTBuffer[i*2+1];

            if(m>0)
            {
                FFTBuffer[i]=10*log10(m);
            }
            else
            {
                FFTBuffer[i]=-180;
            }

            FFTBuffer[i]+=m_DDC1FFTCompensation[i];
        }

        m_DDC1FFTMutex.Unlock();

        m_DDC1Spectrum->SetData(FFTBuffer,FFT_SIZE,m_DDC1Info.SampleRate,(m_DDC1Info.SampleRate-m_DDC1Info.Bandwidth)*0.5);
    }
}

void TDevicePanel::ADCSnapshotCallback(const int16_t *Buffer,uint32_t Count,uint32_t CenterFrequency,uint16_t MaxADCAmplitude)
{
 uint32_t i;
 float m;
 uint32_t AbsoluteFrequencyOfSpectrumCenter;

    for(i=0;i<FFT_SIZE;i++)
    {
        m_ADCFFTBuffer[i*2]=Buffer[i]*m_ADCFFTCoeffs[i];
        m_ADCFFTBuffer[i*2+1]=0;
    }

    FFT(m_ADCFFTBuffer,FFT_SIZE);

    m_ADCFFTMutex.Lock();

    if(m_ADCSpectrumCenterFrequency!=CenterFrequency)
    {
        //compensation data are invalid and they have to be updated

        m_ADCSpectrumCenterFrequency=CenterFrequency;

        AbsoluteFrequencyOfSpectrumCenter=m_RangeCenterFrequency+ADC_SAMPLE_RATE/2/2-CenterFrequency;

        m_DeviceSet->GetSpectrumCompensation(m_DeviceIndex,AbsoluteFrequencyOfSpectrumCenter,ADC_SAMPLE_RATE/2,m_ADCFFTCompensation,FFT_SIZE/2);
    }

    for(i=0;i<FFT_SIZE/2;i++)
    {
        m=m_ADCFFTBuffer[i*2]*m_ADCFFTBuffer[i*2]+m_ADCFFTBuffer[i*2+1]*m_ADCFFTBuffer[i*2+1];

        if(m>0)
        {
            m_ADCFFTBuffer[i]=10*log10(m);
        }
        else
        {
            m_ADCFFTBuffer[i]=-180;
        }

        m_ADCFFTBuffer[i]+=m_ADCFFTCompensation[i];
    }

    m_ADCFFTMutex.Unlock();

    m_ADCSpectrum->SetData(m_ADCFFTBuffer,FFT_SIZE/2,ADC_SAMPLE_RATE/2,CenterFrequency-m_ADCSpectrumBW/2);
}

void TDevicePanel::OnSpectrumTune(TSpectrumTuneEvent &Event)
{
    m_MainWindow->SetDDC1Frequency((uint32_t)Event.GetFrequency());
}

void TDevicePanel::OnNumericEditChanged(wxNotifyEvent &Event)
{
    switch(Event.GetId())
    {
        case ID_PHASE_SHIFT:
            m_DeviceSet->SetDDC1PhaseShift(m_DeviceIndex,m_EPhaseShift->Get()*0.1);
            break;
    }
}

void TDevicePanel::UpdateDeviceState(void)
{
 wxString Str;
 G65DDC_DEVICE_STATE State;
 int Power;

    m_DeviceSet->GetDeviceState(m_DeviceIndex,&State);
    m_DeviceSet->GetPower(&Power);

    if(Power)
    {
        Str=wxEmptyString;

        if(State.Flags & G65DDC_DEVICE_STATE_NO_1PPS)
        {
            Str=wxT("NO 1PPS");
        }

        if(State.Flags & G65DDC_DEVICE_STATE_NOT_ARMED)
        {
            if(!Str.IsEmpty())
            {
                Str+=wxT(" / ");
            }

            Str+=wxT("NOT ARMED");
        }

        if(State.Flags & G65DDC_DEVICE_STATE_TRIGGER_NOT_SENT)
        {
            if(!Str.IsEmpty())
            {
                Str+=wxT(" / ");
            }

            Str+=wxT("TRIGGER NOT SENT");
        }

        if(State.Flags & G65DDC_DEVICE_STATE_NOT_TRIGGERED)
        {
            if(!Str.IsEmpty())
            {
                Str+=wxT(" / ");
            }

            Str+=wxT("NOT TRIGGERED");
        }

        if(State.Flags & G65DDC_DEVICE_STATE_HIGH_TEMPERATURE)
        {
            if(!Str.IsEmpty())
            {
                Str+=wxT(" / ");
            }

            Str+=wxT("HIGH TEMPERATURE");
        }

        if(Str.IsEmpty())
        {
            Str=wxT("OK");
            m_LState->SetForegroundColour(m_LPhaseShift->GetForegroundColour());
        }
        else
        {
            m_LState->SetForegroundColour(wxColor(*wxRED));
        }

        m_LState->SetLabel(Str);

        Layout();
    }
}
