/* ***********************************************************************/
/*                                                                       */
/*                    WiNRADiO G65DDC EXCALIBUR SIGMA                    */
/*                  Header file for G65DDC device driver                 */
/*                      (c) Copyright 2023 RADIXON.                      */
/*                                                                       */
/* ***********************************************************************/
#ifndef __G65DDC_H__
#define __G65DDC_H__

#ifdef HAVE_CONFIG_H
#include <linux/config.h>
#endif

#define G65DDC_USB
#define G65DDC_PCI
//#define G65DDC_DEBUG

#if defined(CONFIG_COMPAT)
#define G65DDC_COMPAT
#endif

#include <linux/version.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/time.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/kthread.h>
#include <linux/timer.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/scatterlist.h>
#include <linux/vmalloc.h>

#ifdef G65DDC_COMPAT
#include <linux/compat.h>
#endif //G65DDC_COMPAT


#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a<<16)|(b<<8)|c)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,8,0)
#pragma GCC error  "Your kernel version is not supported! Kernel v4.8 and above is required."
#endif

#if LINUX_VERSION_CODE>=KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#endif

#include "g65ddc_ioctl.h"

#ifdef G65DDC_COMPAT
#include "g65ddc_ioctl_compat.h"
#endif //G65DDC_COMPAT

#define G65DDC_AUTHOR                   "RADIXON"
#define G65DDC_MODULE_NAME              "g65ddc"
#define G65DDC_MAX_DEVICE_COUNT         32
#define G65DDC_DESCRIPTION              "WiNRADiO G65DDC device driver"


struct g65ddc_io_buffer
{
    struct list_head        list_entry;
    uintptr_t               user_buffer;
    uint32_t                size;
    int                     page_count;
    struct page             **pages;
    bool                    idle;
    int                     status;
    bool                    write;
    uint32_t                bytes_transferred;
    struct scatterlist      *sg_list;
    bool                    changed;
};

struct g65ddc;

typedef void (*g65ddc_fnc_delete)(struct g65ddc *dev);
typedef void (*g65ddc_fnc_uninitialize)(struct g65ddc *dev);
typedef long (*g65ddc_fnc_ioctl)(struct g65ddc *dev,struct file *file, unsigned int cmd,unsigned long arg);
typedef int (*g65ddc_fnc_open)(struct g65ddc *dev);
typedef void (*g65ddc_fnc_close)(struct g65ddc *dev);
typedef int (*g65ddc_fnc_set_power)(struct g65ddc *dev,uint8_t power);
typedef int (*g65ddc_fnc_check_temperature)(struct g65ddc *dev,bool *critical_temperature);
typedef int (*g65ddc_fnc_mmap)(struct g65ddc *dev,struct vm_area_struct *vma);


struct g65ddc
{
    struct device                   dev;
    struct cdev                     cdev;
    uint32_t                        index;
    struct device                   *p_dev;
    struct mutex                    mutex;
    uint8_t                         bus_type;
    uint16_t                        vid_pid[4];
    uint16_t                        fw[4];
    bool                            open;
    union g65ddc_state              state;
    wait_queue_head_t               state_wait_queue;
    struct workqueue_struct         *workqueue;
    struct delayed_work             wd_delayed_work;
    uint32_t                        wd_counter;

    g65ddc_fnc_uninitialize         fnc_uninitialize;
    g65ddc_fnc_delete               fnc_delete;
    g65ddc_fnc_ioctl                fnc_ioctl;

#ifdef G65DDC_COMPAT
    g65ddc_fnc_ioctl                fnc_ioctl_compat;
#endif

    g65ddc_fnc_open                 fnc_open_locked;
    g65ddc_fnc_close                fnc_close_locked;
    g65ddc_fnc_set_power            fnc_set_power_locked;
    g65ddc_fnc_check_temperature    fnc_check_temperature_locked;
    g65ddc_fnc_mmap                 fnc_mmap;
};

#define print_error(__fmt__,...) printk(KERN_ERR G65DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define print_info(__fmt__,...) printk(KERN_INFO G65DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define print_warn(__fmt__,...) printk(KERN_WARNING G65DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)

#define dev_print_error(__dev__,__fmt__,...) printk(KERN_ERR G65DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#define dev_print_info(__dev__,__fmt__,...) printk(KERN_INFO G65DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#define dev_print_warn(__dev__,__fmt__,...) printk(KERN_WARNING G65DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)

#ifdef G65DDC_DEBUG

#define print_dbg(__fmt__,...) printk(KERN_DEBUG G65DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define dev_print_dbg(__dev__,__fmt__,...) printk(KERN_DEBUG G65DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#else

#define print_dbg(__fmt__,...)
#define dev_print_dbg(__dev__,__fmt__,...)

#endif //G65DDC_DEBUG

#ifdef G65DDC_USB

#include <linux/usb.h>

#define G65DDC_USB_TIMEOUT              5000
#define G65DDC_USB_EP_MIN_TRANSFER_SIZE 8192
#define G65DDC_USB_EP_MAX_TRASNFER_SIZE (8*1024*1024)

struct g65ddc_usb;

struct g65ddc_usb_ep
{
    struct g65ddc_usb   *dev;
    uint8_t             address;
    uint16_t            max_packet_size;
    uint8_t             interval;
    uint8_t             type;
    unsigned int        pipe;
    struct usb_anchor   anchor;
    spinlock_t          spinlock;
    struct list_head    buffer_list;
    wait_queue_head_t   wait_queue;
};

struct g65ddc_usb_ep_buffer
{
    struct g65ddc_io_buffer io_buffer;
    struct g65ddc_usb_ep    *ep;
    struct urb              *urb;
};

struct g65ddc_usb
{
    struct g65ddc               base_dev;
    struct usb_device           *usb_dev;
    struct usb_interface        *usb_interface;
    uint32_t                    led_mode;
    struct rw_semaphore         rw_sem;

    unsigned int                snd_ctrl_pipe;
    unsigned int                rcv_ctrl_pipe;

    uint32_t                    ep_count;
    struct g65ddc_usb_ep        *ep;
};

#endif //G65DDC_USB
#ifdef G65DDC_PCI

#include <linux/pci.h>
#include <linux/interrupt.h>

#define DMA_CHANNEL_PTR_COUNT               2
#define NEXT_PTR_INDEX(__ptr_index__)       (1-(__ptr_index__))

#define REG_PCI_INT_MASK                    0x0080
#define REG_PCI_INT_STAT                    0x0084

#define REG_PCI_DMA_ENABLE_IN               0x0400
#define REG_PCI_DMA_ENABLE_OUT              0x0404
#define REG_PCI_DMA_INT_STAT_IN             0x0408
#define REG_PCI_DMA_INT_STAT_OUT            0x040C

#define REG_PCI_TEMP_SENSOR_2               0x0F24
#define REG_PCI_HW_CTRL                     0x0F3C
#define REG_PCI_TEMP_SENSOR_1               0x0F40
#define REG_PCI_TEMP_SENSOR_0               0x0F44

#define REG_PCI_FPGA_CFG_CTRL               0x0F60
#define REG_PCI_FPGA_CFG_DATA_IN            0x0F64

#define REG_PCI_FPGA_2_CTRL                 0x0F70

#define INT_DMA_IN                          0x00000001
#define INT_DMA_OUT                         0x00000002

#pragma pack(push,1)

struct g65ddc_pci_dma_sg_element
{
    uint64_t    address;
    uint64_t    size;
};

#pragma pack(pop)

struct g65ddc_pci_dma_buffer
{
    struct g65ddc_io_buffer             io_buffer;
    int                                 dma_map_sg_count;
    struct list_head                    list_entry_pending;
};

struct g65ddc_pci_dma_ptr
{
    dma_addr_t                          sg_addr_dma;
    void                                *sg_addr_cpu;
    struct g65ddc_pci_dma_buffer        *buffer;
    uint32_t                            reg;
};

struct g65ddc_pci_dma_channel
{
    uint32_t                            index;
    struct g65ddc_pci_dma_ptr           ptr[DMA_CHANNEL_PTR_COUNT];
    uint32_t                            ptr_active_count;
    uint32_t                            ptr_active_index;
    uint32_t                            ptr_idle_index;
    uint32_t                            buffer_size;
    uint32_t                            sg_mem_size;
    bool                                active;
    struct mutex                        mutex;
    struct list_head                    buffer_list;
    struct list_head                    buffer_list_pending;
    wait_queue_head_t                   wait_queue;
    enum dma_data_direction             direction;
};


struct g65ddc_pci
{
    struct g65ddc                       base_dev;
    struct pci_dev                      *pci_dev;
    bool                                fpga;
    bool                                irq_handler_registered;
    uint32_t                            int_mask;
    spinlock_t                          irq_spinlock;
    struct work_struct                  irq_work;

    uint8_t                             *bar2;
    uint32_t                            bar2_size;

    struct rw_semaphore                 rw_sem;

    struct
    {
        uint32_t                        int_stat;
        uint32_t                        int_mask;
        uint32_t                        enabled;
        struct g65ddc_pci_dma_channel   channel[PCI_DMA_CHANNEL_COUNT_FROM_DEVICE];
    }                                   dma_from_device;

    struct
    {
        uint32_t                        int_stat;
        uint32_t                        int_mask;
        uint32_t                        enabled;
        struct g65ddc_pci_dma_channel   channel[PCI_DMA_CHANNEL_COUNT_TO_DEVICE];
    }                                   dma_to_device;
};

#define bar2_read32(__offset__) ioread32(dev->bar2+(__offset__))
#define bar2_write32(__offset__,__value__) iowrite32(__value__,dev->bar2+(__offset__))


#endif //G65DDC_PCI

#endif //__G65DDC_H__
