#!/bin/bash

VERSION="1.05"
DEVICE_NAME="G65DDC"
OS=`uname -s 2>/dev/null`
PLATFORM=`uname -m 2>/dev/null`
MODPROBE=/sbin/modprobe
DEPMOD=/sbin/depmod
LSMOD=/sbin/lsmod
LOG_FILE_NAME="g65ddc.log"
LOG_PATH="../$LOG_FILE_NAME"
MODULE_NAME="g65ddc"
MODULE_FILE_NAME="$MODULE_NAME.ko"
UDEV_RULES_DIR="/etc/udev/rules.d"
UDEV_RULE_FILE="$UDEV_RULES_DIR/WiNRADiO-G65DDC.rules"
UDEVADM="/sbin/udevadm"
API="libg65ddcapi.so"
API_V="$API.$VERSION"
LD=/lib/ld-linux.so.2
LD_x86_64=/lib64/ld-linux-x86-64.so.2
PLATFORMS="i386 i486 i586 i686 x86_64"
LDS="${LD} ${LD} ${LD} ${LD} ${LD_x86_64}"
DIR=$(pwd)

  echo
  echo "**********************************************************************"
  echo "*                  WiNRADiO G65DDC EXCALIBUR SIGMA                   *"
  echo "*                     Driver and API installer                       *"
  echo "**********************************************************************"
  echo

  if [[ $UID -ne 0 ]] ; then
    echo "root privileges are required."
    echo 
    echo "Press ENTER to exit."
    read
    exit 0 
  fi

  if [[ "$OS" != "Linux" ]] ; then
    echo -e "Your operating system is not supported."
    echo
    echo "Press ENTER to exit."
    read
    exit 0
  fi

  SUPPORTED=0

  for i in ${PLATFORMS} ; do
    if [[ "$i" == "$PLATFORM" ]] ; then
        SUPPORTED=1
        break;
    fi
  done

  if [[ "$SUPPORTED" == "0" ]] ; then
    echo -e "Your hardware platform is not supported."
    echo "Supported hardware platforms: $PLATFORMS"
    echo "Detected hardware platform: $PLATFORM"
    echo
    echo "Press ENTER to exit."
    read
    exit 0  
  fi
 
  echo -n -e "Building $DEVICE_NAME kernel module..."

  cd driver

  make clean 2>&1 >"$LOG_PATH"
  make 2>&1 >>"$LOG_PATH"

  if [[ $? -ne 0 ]] ; then
    echo -e "\rFailed to build $DEVICE_NAME kernel module."
    echo "Please, check if kernel development tools are installed"
    echo "and if your kernel supports loadable modules."
    echo
    echo "Compilation log can be found in $LOG_FILE_NAME file."
    echo 
    echo "Press ENTER to exit."
    read
    exit 0
  fi

  if [ ! -f "$MODULE_FILE_NAME" ] ; then
    echo -e "\rFailed to build $DEVICE_NAME kernel module."
    echo "Please, check if kernel development tools are installed"
    echo "and if your kernel supports loadable modules."
    echo
    echo "Compilation log can be found in $LOG_FILE_NAME file."
    echo 
    echo "Press ENTER to exit."
    read
    exit 0
  fi
  
  echo -e "\r$DEVICE_NAME kernel module built successfully."

  echo -n "Creating udev rule..."

  MODULE_LOADED=`$LSMOD | grep $MODULE_NAME | wc -l`  

  if [[ $MODULE_LOADED -ne 0 ]]; then
    $MODPROBE -r $MODULE_NAME
    sleep 2
  fi

  if [ -d $UDEV_RULES_DIR ] ; then
    echo -e "#WiNRADIO $DEVICE_NAME devices\nKERNEL==\"g65ddc*\"\t\t\tMODE=\"0666\"" > "$UDEV_RULE_FILE"

    if [[ -f "$UDEVADM" ]] ; then
      $UDEVADM trigger
    fi

    echo -e "\rUdev rule created successfully."
  else
    echo -e "\r* WARNING: udev rule has not been created. Unable to find udev rules"
    echo "* directory. It is possbile only root will have access to $DEVICE_NAME devices."
  fi


  echo -n "Installing $DEVICE_NAME kernel module..."
  make install 2>&1 >>"$LOG_PATH"

  if [[ $? -ne 0 ]] ; then
    echo -e "\rFailed to install $DEVICE_NAME kernel module."
    echo "Please, check if kernel development tools are installed"
    echo "and if your kernel supports loadable modules."
    echo
    echo "Compilation and installation log can be found in $LOG_FILE_NAME file."
    echo 
    echo "Press ENTER to exit."
    read
    exit 0
  fi

  $DEPMOD -a
  $MODPROBE $MODULE_NAME

  make clean >/dev/null 2>&1 
  rm -f "$LOG_PATH"

  echo -e "\r$DEVICE_NAME kernel module installed successfully."

  cd ..


  echo -e "Installing $DEVICE_NAME API..."

  for p in $PLATFORMS ; do
     f="lib/$p/$API_V"

     if [[ -f "$f" ]] ; then
        printf "%-8s" "$p:"
        status="Not supported"
        lib_dir=
        for l in $LDS ; do
            if test -x $l ; then
                $l --verify "$f" >/dev/null 2>&1
                r=$?
                case $r in
                    0|2)
                        status="Failed"

                        lib_dir=$($l --list "$f" 2>/dev/null | grep /libstdc++.so | sed 's/.*\s\//\//g' | sed 's/\/libstdc++.*//g')

                        if [[ "$lib_dir" != "" ]] ; then
                            if test -f "$lib_dir/$API" ; then
                                 if test -L "$lib_dir/$API" ; then
                                     tmp=$(readlink -f "$lib_dir/$API")

                                     if test -f "$tmp" ; then
                                         rm -f "$tmp" >/dev/null 2>&1
                                     fi
                                 fi
                                 rm -f "$lib_dir/$API" >/dev/null 2>&1
                            fi
                            install -m 755 "$f" "$lib_dir/$API_V" >/dev/null 2>&1

                            if [[ "$?" == "0" ]] ; then
                                cd "$lib_dir"
                                ln -s "$API_V" "$API" >/dev/null 2>&1
                                cd "$DIR"
                                status="OK"
                            fi
                        fi
                        break;
                    ;;
                    *)
                    ;;
                esac
            fi
        done

       printf "%-40s %15s\n" "$lib_dir/$API" "[$status]"
     fi
  done

  echo
  echo "Done."
  echo
  echo "Press ENTER to exit."
  read